
//Refereanser til eksternt grensesnitt/kode som anvendes i denne filen
#include <string>
#include <vector>
#include "Image.h"
#include "Feature.h"
#include "Image_pair.h"
#include "Region.h"
#include "Pixel.h"
#include "Sort.h"
#using <mscorlib.dll>

///Denne koden bruker funksjonalitet definert i std namespace
using namespace std;

///Konstant som representerer uendelig flytverdi for det nevrale nettverket
const float NN_INFINITY = 1000000;


/**
 * @class AVI
 * @brief Hovedklasse i det automatiske visuelle inspeksjonssystemet
 *
 * @version 1.0
 * @author Per Gunnar Brdsen
 */
class AVI
{

	///Dynamisk vektor med bildeobjekter som tilsvarer godkjente bilder
	vector<Image*> approvedImages;

	///Dynamisk vektor med bildeobjekter som tilsvarer underkjente bilder
	vector<Image*> rejectedImages;

	///Dynamisk vektor med egenskapskandidatobjekter
	vector<Feature*> featureCandidates;

	///Dynamisk vektor med pekere til regionobjekter
	vector<Region*> calculatedRegions;

	///Peker til vektor av objekter som holder informasjon om bildepar (denne bindes dynamisk)
	Image_pair *imagePairs;

	///Holder vekten til areal i den lille egenskapsmengden
	float weightArea;

	///Holder vekten til massesenter i den lille egenskapsmengden
	float weightCentroid;

	///Holder vekten til omsluttende rektangel i den lille egenskapsmengden
	float weightBoundingBox;

	///Antall vinduer i bildet
	unsigned short int numberOfWindows;

	///Vektor som holder mappeinformasjon om x-posisjoner (fra bilde til vindu)
	unsigned int *windowVectorMapX;

	///Vektor som holder mappeinformasjon om y-posisjoner (fra bilde til vindu)
	unsigned int *windowVectorMapY;

	///Antall vinduer bildet deles inn i er kvadratet av denne variabelen
	int windowPartitioning;

	///Fargesegmenteringsparameter beskrevet i ColorSegment-klassen
	int minIntensity;

	///Fargesegmenteringsparameter beskrevet i ColorSegment-klassen
	int maxInhomogenity;

	///Fargesegmenteringsparameter beskrevet i ColorSegment-klassen
	int maxBlurIterations;

	///Fargesegmenteringsparameter beskrevet i ColorSegment-klassen
	int minArea;

	///Parameter som er beskrevet i ColorSegment-klassen
	int grow;

	///Parameter som er beskrevet i ColorSegment-klassen
	int writeSegmentationToFile;

	///Peker til sorteringsobjekt (instans av Sort-klassen)
	Sort *sort;

	///Antall egenskapskandidater
	int numberOfFeatureCandidates;

	///Holder navn p fil som gir treningsdata for det nevrale nettverket
	string trainingDataFileName;

	///Holder navn p fil som gir eksekveringsdata for det nevrale nettverket
	string executeDataFileName;

	///Holder navn p fil som informerer om hvilke egenskapskandidater som er egnet
	string selectedFeaturesFileName;

	///Holder navn p katalog til bilder som skal sjekkes
	string folderNameOnline;

	///Holder navn p fil som gir resultatet av eksekvering
	string onlineResultFileName;

	///Dynamisk vektor som holder filnavn p bilder som er godkjente
	vector<string> *approvedNames;
	
	///Dynamisk vektor som holder filnavn p bilder som er underkjente
	vector<string> *rejectedNames;

	///Dynamisk vektor som holder filnavn p bilder som skal sjekkes
	vector<string> *onlineNames;

	///Sti til godkjente bilders posisjon p disk
	string appString;

	///Sti til underkjente bilders posisjon p disk
	string rejString;

	///Sti til diskposisjon av bilder som skal sjekkes
	string onlineString;

	///Nevral nettverk parameter beskrevet i NN_connection-klassen
	int nnNumNeuronsHidden;

	///Nevral nettverk parameter beskrevet i NN_connection-klassen
	int nnMaxEpochs;

	///Nevral nettverk parameter beskrevet i NN_connection-klassen
	int nnEpochsBetweenReports;

	///Nevral nettverk parameter beskrevet i NN_connection-klassen
	float nnDesiredError;

	///Angir summen av alle egnede egenskapskandidaters dimensjoner
	int dimSum;


	/**
	 * @brief Metode som leser inn konfigurasjonsdata fra fil.
	 *
	 * @param *filename Peker til navnet p systemets konfigurasjonsfil.
	 */
	void readConfiguration(const char *fileName);


	/**
	 * @brief Leser en linje fra konfigureringsfilen
	 *
	 * @param input Streng som representerer en linje (med to hovedkraven) i konfigureringsfilen.
	 *
	 * @return En vektor med to verdier som tilsvarer de to hovedkravene for aktuell egenskapskandidat.
	 */
	float * readReq(string input);


	/**
	 * @brief Initialiserer egenskapskandidater
	 *
	 * @param feature Peker til egenskapskandidatobjektet.
	 * @param lines Peker til egenskapskandidatens hovedkrav 1 og 2 fra konfigurasjonsfilen.
	 */
	void initializeFeature(Feature *feature, string *lines);


	/**
	 * @brief Konverterer fra System::String til string
	 *
	 * Denne metoden er kopiert fra MSCE Software Forums, http://www.mcse.ms
	 *
	 * @param str Peker til et System::String-objekt
     *	
	 * @return Inputverdi p string-objekt form.
	 */
	string ToCppString(System::String *str);


	/**
	 * @brief Deler inn bildet i vinduer
	 *
	 * @param dimension Peker til vektor som holder bildets x og y-dimensjon.
	 */
	void windowAssignment(int *dimension);


	/**
	 * @brief Henter aktuelt vindu i bildet.
	 *
	 * @param p Pikselobjekt.
	 * 
	 * @return Det vinduet som den aktuelle pikselen befinner seg i.
	 */
	int getWindow(Pixel p);


	/**
	 * @brief Kalkulerer den lille egenskapsmengden p et bilde.
	 *
	 * @param image Peker til et bildeobjekt.
	 */
	void calculateRegionData(Image *image);


	/**
	 * @brief Lager objekter av bildepar (Image_pair-objekter).
	 *
	 * @return Antall bildeparobjekter.
	 */
	int createImagePairs();


	/**
	 * @brief Beregner den store egenskapskandidatmengden
	 *
	 * @param numberOfPairs Antall bildeparobjekter.
	 */
	void calculateBigFeatureSet(int numberOfPairs);


	/**
	 * @brief Beregner egenskapskandidatverdier for en aktuell region.
	 *
	 * @param region Peker til aktuell region.
	 */
	void calculateRegionFeatureVector(Region *region);


	/**
	 * @brief Beregner egenskapskandidatenes nkkeldata.
	 *
	 * Beregner maskimal og gjennomsnittlig forskjell mellom godkjente bilders objekter seg imellom,
	 * og gjennomsnittlige forskjell mellom godkjente og underkjente bilders objekter.
	 *
	 * @param numberOfPairs Antall bildeparobjekter.
	 */
	void calculateFeatureCandidatesKeyData(int numberOfPairs);


	/**
	 * @brief Henter de egenskapskandidatene som er egnet for bildesettet.
	 *
	 * @return Peker til dynamisk vektor som holder egnede egenskapskandidater
	 */
	vector<int> * selectCandidateFeatures();


	/**
	 * @brief Beregner egnede egenskapskandidater for regioner som ikke er mappet i bildeparene.
	 *
	 * Dette gjres for at alle segmenterte regioner i bildet skal vre med  bidra til 
	 * inputvektoren til det nevral nettverket, som ved "online" eksekvering, og ikke bare de 
	 * regionene som er mappet i bildeparene.
	 *
	 * @param selectedFeatures Peker til dynamisk vektor som holder de egnede egenskapskandidatene.
	 */
	void calculateLeftovers(vector<int> *selectedFeatures);


	/**
	 * @brief Skriver treningsdata til fil.
	 *
	 * @param selectedFeatures Peker til dynamisk vektor som holder de egnede egenskapskandidatene.
	 * @param filename Peker til navnet til fil med treningsdata.
	 *
	 * @return Strrelse til det nevralet netverkets inputvektor.
	 */
	int writeTrainingDataToFile(vector<int> *selectedFeatures, const char *filename);


	/**
	 * @brief Skriver treningsdata til fil. 
	 *
	 * Metode som brukes av writeTrainingDataToFile(...)-metoden.
	 *
	 * @param nnInputVectorSize Strrelse p inputvektor til det nevrale nettverket.
	 * @param nnInputVector Peker til vektor som holder beregnede egenskapskandidatverdier.
	 * @param out Peker til filobjekt hvor utskrift pgr.
	 */
	void writeTrainingVector(unsigned int nnInputVectorSize, float *nnInputVector, FILE *out);


	/**
	 * @brief Oppdaterer inputvektor for det nevrale nettverket med beregnede egenskapskandidatverdier.
	 *
	 * @param start Indeks som angir aktuell startposisjon i nnInputVector.
	 * @param nnInputVector Peker til inputvektor for det nevrale nettverket.
	 * @param region Peker til aktuell region som brukes til  finne beregnede verdier for egenskapskandidater.
	 * @param selectedFeatures Peker til dynamisk vektor som holder de egnede egenskapskandidatene.
	 */
	void insertValuesIntoVector(int start, float *nnInputVector, Region *region, vector<int> *selectedFeatures);


	/**
	 * @brief Henter ut en liste av egnede egenskapskandidater fra fil.
	 *
	 * @param fileName Peker til navn p fil som angir hvilke egenskapskandidater som er egnede.
     *	
	 * @return Peker til en dynamisk vektor med pekere til egnede egenskapskandidatobjekter.
	 */
	vector<Feature*> * readSelectedFeaturesFromFile(const char *fileName);


	/**
	 * @brief Henter ut det aktuelle egenskapskandidatobjektet tilsvarende identifikasjon gitt i fil.
	 *
	 * @param id Identifikasjon p aktuell egenskapskandidat fra fil.
     *	
	 * @return Peker til det aktuelle egenskapskandidatobjektet.
	 */
	Feature * getFeatureFromID(int id);


	/**
	 * @brief Beregner inputvektor til det nevrale nettverket i eksekveringsfase.
	 *
	 * @param selectedFeatures Peker til dynamisk vektor som holder de egnede egenskapskandidatene.
	 * @param regions Peker til en dynamisk vektor som holder pekere til regionobjekter.
     *	
	 * @return Peker til vektor som holder inputvektor til det nevrale nettverket.
	 */
	float * calculateOnlineImageData(vector<Feature*> *selectedFeatures, vector<Region*> *regions);


	/**
	 * @brief Avrunder flyttall med nedre og vre grense.
	 *
	 * @param x Verdi som skal avrundes.
     *	
	 * @return Avrundet verdi av flyttallet.
	 */
	float roundOf(float x);


public:

	/**
	 * @brief Konstruktor
	 *
	 * @param online Parameter som angir om systemet skal kjre i offline eller online (med verdier henholdsvis 0 og 1).
	 * @param configurationFileName Peker til navnet p systemets konfigurasjonsfil.
	 */
	AVI(int online, const char *configurationFileName);

	/**
	 * @brief Destruktor
	 */
	~AVI();

};