
//Refereanser til eksternt grensesnitt/kode som anvendes i denne filen
#include <string>
#include <cmath>
#include <ctime>
#include <vector>
#include "ColorSegment.h"
#include "Region.h"
#include "Pixel.h"
#include "ipl98\cpp\image.h"

//Denne kilden bruker kode definert i navnerommet std
using namespace std;

//Definerer konstanter
const int GRAYSCALE = 256;
const int GRAYSCALE_SQUARE = GRAYSCALE * GRAYSCALE;
const int INFINITY = GRAYSCALE * GRAYSCALE_SQUARE;
const int LITTLE_MASK = 3;
const int BIG_MASK = 5;
const int SIDE = (BIG_MASK - 1) / 2;
const int UNMARKED = -1;



/**
 * @brief Normaliserer en piksel i bildet.
 *
 * Metoden skalerer opp fargekomponentenes intensitetet, slik at pikselens hyeste 
 * fargekomponent har maksimal intensitet (255). Piksler med lavere intensitet enn 
 * terskelkravet nulles ut (sette til svart).
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param i Angir aktuell piksel i bildet som indeks i den rde komponenten av bildevektor p.
 * @param s Bildets strrelse (bredde * hyde).
 * @param r Verdi i pikselens rde komponent.
 * @param g Verdi i pikselens grnne komponent.
 * @param b Verdi i pikselens bl komponent.
 * @param t Terskel som angir et minstekrav til fargeintensitet. Piksler med intensiteter 
 *	        (r*r+g*g+b*b) mindre enn denne terskelen settes til svart (r=0, g=0, b=0).
 */
void normalize(int *p, int i, int s, int r, int g, int b, int t)
{
	int is = i + s;
	if ( r*r + g*g + b*b > t )
	{
		int m = max(r,g);
		m = max(b,m);
		float f = (GRAYSCALE-1) / (float) m;
		p[i] = (int) (r*f);
		p[is] = (int) (g*f);
		p[is+s] = (int) (b*f);
	}
	else
	{
		p[i] = 0;
		p[is] = 0;
		p[is+s] = 0;
	}
}



/**
 * @brief Sjekker om en piksel ligger p en av bildets fire kanter.
 *
 * @param i Angir aktuell piksel i bildet.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 *
 * @return True dersom aktuell piskel ligger p kanten av bildet, og false ellers.
 */
bool imageBorderPixel(int i, int w, int h)
{
	int y = i / w;
	int x = i % w;
	if ( y<SIDE || x<SIDE || y>h-SIDE || x>w-SIDE ) return true;
	return false;
}



/**
 * @brief Beregner strste "varians" til aktuell piksel innenfor en valgt maskestrrelse.
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param s Bildets strrelse (bredde * hyde).
 * @param r Angir aktuell piksel i bildet som indeks i den rde komponenten av bildevektor p.
 * @param m Angir maskestrrelse.
 *
 * @return Den strste kvadrerte sum av fargekomponentenes "varians" mellom aktuell piksel 
 *		   og naboverdiene innenfor valgt maskestrrelse.
 */
int variance(int *p, int s, int r, int m)
{
	int v = 0;
	int j = (m * m - 1) / 2;
	int g = r + s;
	int b = g + s;
	for (int a=-j; a<=j; a++)
	{
		if ( a==0 ) continue;
		int d = (int) ( pow(p[r] - p[r+a], 2) + pow(p[g] - p[g+a], 2) + pow(p[b] - p[b+a], 2) );
		v = max(d,v);
	}
	return v;
}



/**
 * @brief Sjekker om aktuell piksel er posisjonert p regionens kontur eller innenfor.
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param s Bildets strrelse (bredde * hyde).
 * @param i Angir aktuell piksel i bildet som indeks i den rde komponenten av bildevektor p.
 * @param t Terskel som angir maksimalt tillatte inhomogenitet i regioner.
 *
 * @return True dersom aktuell piskel ligger p en regionenkontur, og false ellers.
 */
bool edge(int *p, int s, int i, int t)
{
	int v1 = variance(p, s, i, LITTLE_MASK);
	int v2 = variance(p, s, i, BIG_MASK);
	if ( v2-v1 <= t ) return false;
	return true;
}



/**
 * @brief Oppdaterer bildevektoren med glattet verdi av aktuell piksel basert p 
 *		  gjennomsnittet av de tte nrmeste nabopikslenes verdier.
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param s Bildets strrelse (bredde * hyde).
 * @param r Angir aktuell piksel i bildet som indeks i den rde komponenten av bildevektor p
 */
void average(int *p, int s, int r)
{
	int sumR = 0;
	int sumG = 0;
	int sumB = 0;
	int g = r + s;
	int b = g + s;
	for (int i=-4; i<=4; i++)
	{
		if ( i==0 ) continue;
		sumR += p[r+i];
		sumG += p[g+i];
		sumB += p[b+i];
	}
	p[r] = sumR / 8;
	p[g] = sumG / 8;
	p[b] = sumB / 8;
}



/**
 * @brief Metode som glatter bildet i posisjoner som tilsvarer regioners innhold 
 *        (dvs ikke langs konturer).
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 * @param t1 Terskel som angir maksimalt tillatte inhomogenitet i regioner.
 * @param t2 Terskel som angir maksimalt antall iterasjoner som glatter bildet.
 */
void blur(int *p, int w, int h, int t1, int t2)
{
	int s = w * h;
	for (int i=0;i<t2;i++)
	{
		bool stop = true;	
		for (int r=0; r<s; r++)
		{
			if ( imageBorderPixel(r, w, h) ) continue;
			if ( !edge(p, s, r, t1) )
			{
				average(p, s, r);
				stop = false;
			}	
		}
		if ( stop ) return;
	}
}



/**
 * @brief Lager en vektor med informasjon over hvilke piksler og regioner som 
 *        hrer sammen.
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 * @param t En terskel som angir maksimalt tillatte inhomogenitet i regioner.
 *
 * @return Peker til vektor som markerer hvilken regioner pikslene tilhrer.
 */
int * label(int *p, int w, int h, int t)
{
	int s = w * h;
	int *label = new int[s];
	int *east = new int[s];
	int *south = new int[s];
	int id = 0;
	vector<int> *region = new vector<int>;
	bool *temp = new bool[s];
	for (int i=0; i<s; i++)
	{
		int is = i+s;
		int iss = is+s;
		east[i]=0;
		south[i]=0;
		if ( i%w < w-1 ) { east[i] = abs( p[i] - p[i+1] ) + abs( p[is] - p[is+1] ) + abs( p[iss] - p[iss+1] ); }
		if ( i/w < h-1 ) { south[i] = abs( p[i] - p[i+w] ) +  abs( p[is] - p[is+w] ) + abs( p[iss] - p[iss+w] ); }
		label[i] = UNMARKED;
		temp[i] = false;
	}
	for (int i=0; i<s; i++)
	{
		if ( label[i]==UNMARKED )
		{
			region->push_back(i);
			while ( region->size() > 0 )
			{
				int j = region->at(0); 
				region->erase( region->begin() );
				label[j] = id;
				int x=j%w, y=j/w, N=j-w, E=j+1, S=j+w, W=j-1;
				int NE=N+1, SE=S+1, SW=S-1, NW=N-1;
				if ( y>0 && label[N]==UNMARKED && south[N]<=t && !temp[N] ) 
				{
					region->push_back(N);
					temp[N] = true;
				}
				if ( x<w-1 && label[E]==UNMARKED && east[j]<=t && !temp[E] ) 
				{
					region->push_back(E);
					temp[E] = true;
				}
				if ( y<h-1 && label[S]==UNMARKED && south[j]<=t && !temp[S] )
				{
					region->push_back(S);
					temp[S] = true;
				}
				if ( x>0 && label[W]==UNMARKED && east[W]<=t && !temp[W] )
				{
					region->push_back(W);
					temp[W] = true;
				}
			}
			id++;
		}
	}
	delete[] temp, region;
	return label;
}



/**
 * @brief Teller opp antall ulike regioner i bildet.
 *
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param s Bildets strrelse (bredde * hyde).
 *
 * @return Antall regioner i bildet.
 */
int countNumberOfLabels(int *l, int s)
{
	int m = 0;
	for (int i=0; i<s; i++)
	{
		m = max(m, l[i]);
	}
	return m + 1;
}



/**
 * @brief Metode som teller opp antall piksler, samt summerer opp verdier av den 
 *        rde, grnne og bl komponenten over alle pikslene i hver region.
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param numberOfLabels Antall regioner i bildet. 
 * @param s Bildets strrelse (bredde * hyde).
 *
 * @return Peker til en vektor som videre inneholder pekere til to vektorer. Den frste 
 *         av disse inneholder komponentvis sum av fargeverdiene i hver region, og den
 *         andre inneholder antall piksler i hver region.
 */
int ** calculateRegion(int *p, int *l, int numberOfLabels, int s)
{
	int **v = new int*[2];
	v[0] = new int[ 3 * numberOfLabels ];
	v[1] = new int[ numberOfLabels ];
	for (int i=0; i<numberOfLabels; i++)
	{
		int ir = i+numberOfLabels;
		v[0][i] = 0;
		v[0][ir] = 0;
		v[0][ir+numberOfLabels] = 0;
		v[1][i] = 0;
	}
	for (int i=0; i<s; i++)
	{
		int is = i + s;
		int lin = l[i] + numberOfLabels;
		int linn = lin + numberOfLabels;
		v[ 0 ][ l[i] ] += p[ i ];
		v[ 0 ][ lin ] += p[ is ];
		v[ 0 ][ linn ] += p[ is + s ];
		v[ 1 ][ l[i] ]++;
	}
	return v;
}



/** 
 * @brief Sjekker om aktuell piksel ligger p konturen til en svart/uspesifisert
 *		  region.
 *
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param i Angir aktuell piksel i bildet.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 *
 * @return True dersom aktuell piksel ligger p konturen til en svart/uspesifisert
 *		   region, og false ellers.		 
 */
bool blackBorder(int *l, int i, int w, int h)
{
	int x=i%w, y=i/w, N=i-w, E=i+1, S=i+w, W=i-1;
	if ( y>0 && l[N]!=UNMARKED ) return true;
	if ( x<w-1 && l[E]!=UNMARKED ) return true;
	if ( y<h-1 && l[S]!=UNMARKED ) return true;
	if ( x>0 && l[W]!=UNMARKED ) return true;
	return false;
}



/**
 * @brief Eliminerer bort regioner som ikke oppfyller minstekravet til areal.
 *
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 * @param count Peker til vektor som holder oversikt over antall piksler i hver region.
 * @param minArea Nedre grense for hvor lite areal akseptable regioner m ha.
 *
 * @return Peker til en dynamisk vektor som inneholder indeks til piksler som ligger
 *		   p konturen av svarte/uspesifiserte regioner.
 */
vector<int> * eliminateSmallRegions(int *l, int w, int h, int *count, int minArea)
{
	vector<int> *b = new vector<int>;
	for (int i=0; i<w*h; i++)
	{
		if ( count[ l[i] ] < minArea )
		{
			l[i] = UNMARKED;
			b->push_back(i);
		}
	}
	for (unsigned int i=0; i<b->size(); i++)
	{
		int j = b->at(i);
		if ( !blackBorder(l, j, w, h) ) b->erase( b->begin() + i );	
	}
	return b;
}



/**
 * @brief Finner hvilken naboregion som er best skikket til  spise opp aktuell piksel.
 *
 * @param i Angir aktuell piksel i bildet som indeks i den rde komponenten av bildevektor p.
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 * @param values Peker til vektor som holder oversikt over komponentvis sum av fargeverdier
 *				 i hver region.
 * @param count Peker til vektor som holder oversikt over antall piksler i hver region.
 * @param regions Antall regioner i bildet.
 *
 * @return Indeks til en piksel i bildet, som har den tilhrende regionen som er best egnet
 *		   til  spise opp den svarte/uspesifiserte pikselen.
 */
int bestNeigbourRegion(int i, int *p, int *l, int w, int h, int *values, int *count, int regions)
{
	int m=INFINITY, index=-1, x=i%w, y=i/w, s=w*h, is=i+s, iss=is+s;
	int N=i-w, E=i+1, S=i+w, W=i-1;
	if ( y>0 && l[N]!=UNMARKED )
	{
		int lnr = l[N] + regions;
		float f = (float) ( 1 / count[ l[N] ] ); 
		int r = (int) ( f * values[ l[N] ] );
		int g = (int) ( f * values[ lnr ] );
		int b = (int) ( f * values[ lnr + regions ] );
		m = abs( p[i] - r ) + abs( p[is] - g ) + abs( p[iss] - b );
		index = N;
	}
	if ( x<w-1 && l[E]!=UNMARKED )
	{
		int ler = l[E] + regions;
		float f = (float) ( 1 / count[ l[E] ] ); 
		int r = (int) ( f * values[ l[E] ] );
		int g = (int) ( f * values[ ler ] );
		int b = (int) ( f * values[ ler + regions ] );
		int v = abs( p[i] - r ) + abs( p[is] - g ) + abs( p[iss] - b );
		if ( v < m ) { m=v; index=E; }
	}
	if ( y<h-1 && l[S]!=UNMARKED )
	{
		int lsr = l[S] + regions;
		float f = (float) ( 1 / count[ l[S] ] ); 
		int r = (int) ( f * values[ l[S] ] );
		int g = (int) ( f * values[ lsr ] );
		int b = (int) ( f * values[ lsr + regions ] );
		int v = abs( p[i] - r ) + abs( p[is] - g ) + abs( p[iss] - b );
		if ( v < m ) { m=v; index=S; }
	}
	if ( x>0 && l[W]!=UNMARKED )
	{
		int lwr = l[W] + regions;
		float f = (float) ( 1 / count[ l[W] ] ); 
		int r = (int) ( f * values[ l[W] ] );
		int g = (int) ( f * values[ lwr ] );
		int b = (int) ( f * values[ lwr + regions ] );
		int v = abs( p[i] - r ) + abs( p[is] - g ) + abs( p[iss] - b );
		if ( v < m ) { m=v; index=W; }
	}
	return index;
}



/**
 * @brief Fusjonerer aktuell piksel til den beste naboregionen som innfrir arealkravet.
 *
 * @param i Angir indeks til aktuell piksel i bildet.
 * @param j Angir indeks til en piksel med tilhrende region som skal spise opp den
 *          svarte/uspesifiserte pikselen.
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param b Peker til en dynamisk vektor som inneholder indeks til piksler som ligger
 *          p konturen av svarte/uspesifiserte regioner.
 * @param values Peker til vektor som holder oversikt over komponentvis sum av fargeverdier
 *				 i hver region.
 * @param count Peker til vektor som holder oversikt over antall piksler i hver region.
 * @param regions Antall regioner i bildet.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 */
void merge(int i, int j, int *p, int *l, vector<int> *b, int *values, int *count, int regions, int w, int h)
{
	int s = w*h;
	int jr = l[j]+regions;
	int is = i+s;
	values[ l[j] ] += p[i];
	values[ jr ] += p[is];
	values[ jr+regions ] += p[is+s];
	count[ l[j] ]++;
	int x=i%w, y=i/w, N=i-w, E=i+1, S=i+w, W=i-1;
	if ( y>0 && l[N]==UNMARKED && !blackBorder(l,N,w,h) ) b->push_back(N);
	if ( x<w-1 && l[E]==UNMARKED && !blackBorder(l,E,w,h) ) b->push_back(E);
	if ( y<h-1 && l[S]==UNMARKED && !blackBorder(l,S,w,h) ) b->push_back(S);
	if ( x>0 && l[W]==UNMARKED && !blackBorder(l,W,w,h) ) b->push_back(W);
	l[i] = l[j];
}



/**
 * @brief Lar regioner spise opp sm piksler tilhrende regioner som ikke 
 *        tilfredstiller arealkravet.
 *
 * @param b Peker til en dynamisk vektor som inneholder indeks til piksler som ligger
 *          p konturen av svarte/uspesifiserte regioner.
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 * @param values Peker til vektor som holder oversikt over komponentvis sum av fargeverdier
 *				 i hver region.
 * @param count Peker til vektor som holder oversikt over antall piksler i hver region.
 * @param regions Antall regioner i bildet.
 */
void regionGrow(vector<int> *b, int *p, int *l, int w, int h, int *values, int *count, int regions)
{
	while ( b->size() > 0 )
	{
		int i = b->at(0);
		b->erase( b->begin() );
		int j = bestNeigbourRegion(i, p, l, w, h, values, count, regions);
		if ( j > -1 ) {	merge(i, j, p, l, b, values, count, regions, w, h);	}
    }
}



/**
 * @brief Oppretter en vektor som mapper regioner med tilfeldige farger.
 *
 * @param numberOfLabels Antall regioner i bildet.
 *
 * @return Peker til en vektor som mapper regioner til farger.
 */
int * createRandomColorMap(int numberOfLabels)
{
	int *map = new int[3*numberOfLabels];
	srand ( time(NULL) );
	for (int i=0; i<numberOfLabels; i++)
	{
		int in = i + numberOfLabels;
		int LIM = GRAYSCALE - 1;
		map[i] = (int) ( LIM * rand() / RAND_MAX );
		map[in] = (int) ( LIM * rand() / RAND_MAX );
		map[in+numberOfLabels] = (int) ( LIM * rand() / RAND_MAX );
	}
	return map;
}



/**
 * @brief Oppdaterer bildet slik at pikslenes fargeverdier settes i henhold til hvilke regioner de tilhrer.
 *
 * @param p Peker til vektor av pikselverdier over alle tre fargekomponenter.
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param map Peker til en vektor som mapper regioner til farger.
 * @param s Bildets strrelse (bredde * hyde).
 * @param numberOfLabels Antall regioner i bildet.
 */
void mapColorToLabels(int *p, int *l, int *map, int s, int numberOfLabels)
{
	for (int i=0; i<s; i++)
	{
		int is = i + s;
		int lis = l[i] + numberOfLabels;
		if ( l[i] == UNMARKED ) 
		{
			p[i] = 0;
			p[is] = 0;
			p[is+s] = 0;
		}
		else
		{
			p[i] = map[ l[i] ];
			p[is] = map[ lis ];
			p[is+s] = map[ lis + numberOfLabels ];
		}
	}
}



/**
 * @brief Sjekker om aktuell piksel er kontur i en region eller ikke.
 *
 * @param l Peker til vektor som holder oversikt over hvilke piksler og regioner
 *          som hrer sammen.
 * @param x Angir x-akseverdien til aktuell piksel i bildet.
 * @param y Angir y-akseverdien til aktuell piksel i bildet.
 * @param w Bildets bredde.
 * @param h Bildets hyde.
 * @param i Angir aktuell piksel i bildet.
 *
 * @return True dersom aktuell piksel ligger p en regionkontur, false ellers.
 */
bool regionBorder(int *l, int x, int y, int w, int h, int i)
{
	if ( y>0 && l[i-w]!=l[i] ) return true;
	if ( x<w-1 && l[i+1]!=l[i] ) return true;
	if ( y<h-1 && l[i+w]!=l[i] ) return true;
	if ( x>0 && l[i-1]!=l[i] ) return true;
	return false;
}



/**
 * @brief Destruktor
 */
ColorSegment::~ColorSegment()
{
	//Tom destruktor
}



/**
 * @brief Konstruktor
 *
 * @param _minIntensity En terskel som angir et minstekrav til fargeintensitet. Piksler med 
 *				   	    intensiteter (r*r+g*g+b*b) mindre enn denne terskelen vil ikke bidra
 *						i segmenteringsprosessen.
 * @param _maxInhomogenity En terskel som angir maksimalt tillatte inhomogenitet i regioner. 
 *						   Dvs at pikslene P1 og P2 bare sles sammen til en region dersom 
 *						   |r1-r2| + |g1-g2| + |b1-b2| <= _maxInhomogenity, der r1 og r2 er 
 *						   den rde intensiteten til henholdsvis piksel P1 og P2 osv.
 * @param _maxBlurIterations vre grense for antall iterasjoner som glatter bildet.
 * @param _minArea Nedre grense for areal med tanke p akseptable regionsegmenteringer.
 * @param _grow Verdi som angir om region growing skal igangsettes for  spise opp usegmenterte
 *			    piksler eller ikke.
 * @param _writeToFile Verdi som angir om segmenteringsresultatene skal skrives ut til fil 
					eller ikke.
*/
ColorSegment::ColorSegment(int _minIntensity, int _maxInhomogenity, int _maxBlurIterations, 
						   int _minArea, int _grow, int _writeToFile)
{
	minIntensity = _minIntensity;
	maxInhomogenity = _maxInhomogenity;
	maxBlurIterations = _maxBlurIterations;
	minArea = _minArea;
	grow = _grow;
	writeToFile = _writeToFile;
}



/**
 * @brief Metode som henter dimensjonen til et bilde
 *
 * @param filename Navn p inputbilde
 *
 * @return Peker til en vektor av lengde 2, der frste og andre rad henholdsvis inneholder
 *		  bildets bredde og hyde.
 */
int * ColorSegment::getDimension(const char *filename)
{
	ipl::CImage img;
	img.Load(filename);    
	int *dim = new int[2];
	dim[0] = img.GetWidth();
	dim[1] = img.GetHeight();
	return dim;
}



/**
 * @brief Metode som segmenterer og returnerer et image-objekt av et bilde.
 * 
 * @param filename Navn p bilde som skal segmenteres
 * @param index Lengde p stien (path) hvor det aktuelle bildet ligger lagret
 *
 * @return Image objekt av bildet.
 */
Image * ColorSegment::segment(const char * filename, int index)
{
	ipl::CImage img;
	img.Load(filename);    
	const int w = img.GetWidth();
	const int h = img.GetHeight();
	int s = w * h;
	int *p = new int[3*s];
	int t = 3 * minIntensity * minIntensity;
	for (int y=0; y<h; y++)
	{
		for (int x=0; x<w; x++)
		{
			int r = img.GetRedComponentFast(x, y);
			int g = img.GetGreenComponentFast(x, y);
			int b = img.GetBlueComponentFast(x, y);
			normalize(p, y*w+x, s, r, g, b, t);
		}
	}

	blur(p, w, h, maxInhomogenity, maxBlurIterations);
	int *l = label(p, w, h, maxInhomogenity);
	int numberOfLabels = countNumberOfLabels(l, s);
	int **regionData = calculateRegion(p, l, numberOfLabels, s);
	vector<int> *b = eliminateSmallRegions(l, w, h, regionData[1], minArea);
	if ( grow>0 && b->size()>0 ) regionGrow(b, p, l, w, h, regionData[0], regionData[1], numberOfLabels);

	vector<Pixel> *region = new vector<Pixel>[s];
	vector<Pixel> *border = new vector<Pixel>[s];
	int *rSum = new int[s];
	int *gSum = new int[s];
	int *bSum = new int[s];
	for (int i=0; i<s; i++)
	{
		rSum[i] = 0;
		gSum[i] = 0;
		bSum[i] = 0;
	}
	vector<int> map;
	for (int y=0; y<h; y++)
	{
		for (int x=0; x<w; x++)
		{
			int index = y*w + x;
			int indexs = index + s;
			int indexss = indexs + s;
			Pixel pix(x, y, p[index], p[indexs], p[indexss] );
			int i = l[index];
			if ( regionBorder(l,x,y,w,h,index) ) border[i].push_back(pix);
			if ( i<0 ) continue;
			if ( region[i].size() == 0 ) map.push_back(i);
			region[i].push_back(pix);
			rSum[i] += pix.r;
			gSum[i] += pix.g;
			bSum[i] += pix.b;
		}
	}

	vector<Region*> *regions = new vector<Region*>;
	for (unsigned int i=0; i<map.size();i++)
	{
		int j = map.at(i);
		Region *r = new Region( region[j], border[j], rSum[j], gSum[j], bSum[j] );
		regions->push_back(r);
	}

	Image *im = new Image( regions, filename );

	if ( writeToFile > 0 )
	{
		int *map = createRandomColorMap(numberOfLabels);
		mapColorToLabels(p, l, map, s, numberOfLabels);
		for (int y=0; y<h; y++)
		{
			for (int x=0; x<w; x++)
			{
				int i = y*w + x;
				int is = i + s;
				if ( l[index]<0 )
				{
					img.SetRedComponentFast(x, y, 0);
					img.SetGreenComponentFast(x, y, 0);
					img.SetBlueComponentFast(x, y, 0);
				}
				else
				{
					img.SetRedComponentFast(x, y, p[i]);
					img.SetGreenComponentFast(x, y, p[is]);
					img.SetBlueComponentFast(x, y, p[is+s]);
				}
			}
		}
		string st(filename);
		st.insert(index,"_");
		img.Save( st.c_str() );
	}

	delete[] p, l, b, map;
	delete[] regionData[0], regionData[1], regionData;
	delete[] rSum, gSum, bSum;

	return im;

}