#pragma once

//Refereanser til eksternt grensesnitt/kode som anvendes i denne filen
#include "Region.h"

///Konstant som representerer uendelig
const float INFINITY = 1000000;


/**
 * @class Feature
 * @brief Abstrakt klasse som inneholder felles funksjonalitet for ulike egenskapskandidater.
 *
 * @version 1.0
 * @author Per Gunnar Brdsen
 */
class Feature
{

protected:
	
	///Egenskapskandidatens identifikasjon.
	unsigned short int ID;

	///Egenskapskandidatens dimensjonkrav.
	unsigned short int DIM;

	///Egenskapskandidatens navn
	const char *NAME;

	///Egenskapskandidatens poengsum.
	unsigned short int score;

	///Peker til vektor som holder hovedkrav 1 (nedre grense for likhet mellom godkjente bilders objekter seg imellom).
	float *req1;

	///Peker til vektor som holder hovedkrav 2 (nedre grense for ulikhet mellom godkjente og underkjente bilders objekter).
	float *req2;

	///Peker til vektor som holder maksimal forskjell mellom godkjente bilders objekter.
	float *homogenMax;

	///Peker til vektor som holder gjennomsnittlig forskjell mellom godkjente bilders objekter.
	float *homogenMean;

	///Peker til vektor som holder gjennomsnittlig forskjell mellom godkjente og underkjente bilders objekter.
	float *inhomogenMean;

	/** 
	 * @brief Finner absoluttverdien til aktuelle flyttal.
	 *
	 * @param value Flyttallet man nsker absoluttverdien av.
	 *
	 * @return Flyttallets absouluttverdi.
	 */
	float absoluteValue(float value);


	/**
	 * @brief Garanterer verdier mellom en nedre og vre definert grense.
	 * 
	 * @param Flyttallet som skal hndteres.
	 *
	 * @return Verdi innenfor definert omrde.
	 */
	float valueHandler(float value);


public:

	/** 
	 * @brief Konstruktor
	 */
	Feature(void);


	/** 
	 * @brief Destruktor
	 */
	~Feature(void);


	/** 
	 * @brief Initialiserer egenskapskandidaten.
	 *
	 * @param abc Dobbel pekerstruktur (vektor v av 2-dimensjonale vektorer w), der 
	 *	          vektor v sin dimensjon tilsvarer antall dimensjoner i egenskapskandidaten,
	 *		      og vektorene w hver har hovedkrav 1 og 2 som komponentverdier.
	 */
	void initialize(float **abc);


	/** 
	 * @brief Henter egenskapskandidatens identifikasjonen.
	 *
	 * @return Egenskapskandidatens identifikasjonen.
	 */
	int getId();


	/** 
	 * @brief Henter egenskapskandidatens navn.
	 *
	 * @return Egenskapskandidatens navn.
	 */
	const char * getName();


	/** 
	 * @brief Henter egenskapskandidatens krav til dimensjon.
	 *
	 * @return Egenskapskandidatens dimensjon.
	 */
	int getDim();


	/** 
	 * @brief Henter egenskapskandidatens poengsum
	 *
	 * @return Egenskapskandidatens poengsum
	 */
	int getCredit();
	

	/** 
	 * @brief Beregner egenskapskandidaten p aktuell region.
	 *
	 * Pure virtual function som m implementeres av de forskjellige egenskapskandidatene
	 * som arver funksjonalitet fra denne klassen.
	 *
	 * @param r Peker til aktuell region som egenskapskandidaten skal beregnes for.
	 *
	 * @return Peker til en vektor av flytverdier.
	 */
	virtual float * calculate(Region *r) = 0;


	/** 
	 * @brief Beregner differansen av egenskapskandidatens kalkulerte resultater av 
	 *        to aktuelle regioner.
	 *
	 * @param r1 Peker til den ene regionen.
	 * @param r2 Peker til den andre regionen.
	 *
	 * @return Peker til en vektor av flytverdier av ulikhet mellom de to regionene.
	 */
	float * difference(Region *r1, Region *r2);


	/** 
	 * @brief Adderer egenskapskandidatens differanseverdier av to regioner til variablene 
	 *		  homogenMean eller inhomogenMean, for  senere kunne beregne
	 *		  gjennomsnittlige beregninger.
	 *
	 * @param homogen Boolsk verdi som ved (true) forteller om det er to godkjente 
     *                bilders objekter som vurderes, og som ved (false) forteller
	 *				  at det er objekter av et gokjent og et underkjent bilde som
	 *				  vurders.
	 * @param diff Peker til differansen mellom denne egenskapskandidatens beregninger
	 *			   for to regioner.
	 */
	void add(bool homogen, float *diff);


	/** 
	 * @brief Oppdaterer pekeren som holder oversikt over maksimal forskjell mellom 
	 *		  objekter av godkjente bilders regioner seg imellom.
	 *
	 * @param diff Peker til differansen mellom denne egenskapskandidatens beregninger
	 *			   for to regioner.
	 */
	void maxHomogen(float *diff);


	/** 
	 * @brief Beregner gjennomsnittlig regiondifferanseverdi for denne egenskapskandidaten.
	 *
	 * @param homogen Boolsk verdi som ved (true) forteller om det er to godkjente 
     *                bilders objekter som vurderes, og som ved (false) forteller
	 *				  at det er objekter av et gokjent og et underkjent bilde som
	 *				  vurders.
	 * @param divisor Divisor for  rekne gjennomsnittlige differanseverdier for denne
	 *				  egenskapskandidaten.
	 */
	void divide(bool homogen, int divisor);

};