#pragma once

//Refereanser til eksternt grensesnitt/kode som anvendes i denne filen
#include "Image.h"
#include "Sort.h"


/**
 * @brief Kvadrerer en heltallsverdi.
 *
 * @param x Heltallsverdi som skal kvadreres.
 *
 * @return Kvadratet av heltallsverdien.
 */
int kvadrat(int x); 


/**
 * @brief Kvadratroten av en heltallsverdi.
 *
 * @param x Heltallsverdi som kvadratroten nskes funnet av.
 *
 * @return Kvadratroten av heltallsverdien.
 */
float rot(int x);


/**
 * @class Image_pair
 * @brief Klasse som holder informasjon om et bildepar.
 *
 * @version 1.0
 * @author Per Gunnar Brdsen
 */
class Image_pair
{

	///Det ene bildet i bildeparet
	Image *image1;

	///Det andre bildet i bildeparet
	Image *image2;

	///Boolsk verdi som forteller om bildeparet bestr av to godkjente bilder (true), eller ett godkjent og ett underkjent bilde (false)
	bool homogenious;

	///Peker til en vektor som inneholder vektingen av den lille egenskapsmengden (areal, massesenter og omlsuttende rektangel).
	float *weight;

	///Peker til en vektor som holder oversikt over den beste mappingen/korrepondansen mellom regioner i bildeparet
	int *bestCorrespondenceVector;

public:

	/**
	 * @brief Konstruktor
	 */
	Image_pair();


	/**
	 * @brief Konstruktor
	 *
	 * @param _image1 Peker til den ene bildet i bildeparet.
	 * @param _image2 Peker til det andre bildet i bildeparet.
	 * @param _homogenious Boolsk verdi som forteller om bildeparet bestr av to godkjente 
	 *					   bilder (true), eller ett godkjent og ett underkjent bilde (false).
	 * @param weight Peker til en vektor som inneholder vektingen av den lille 
	 *				 egenskapsmengden (areal, massesenter og omlsuttende rektangel).
	 * @param sort Peker til sorteringsobjektet sort.
	 */
	Image_pair(Image *_image1, Image *_image2, bool _homogenious, float *weight, Sort *sort);


	/**
	 * @brief Destruktor
	 */
	~Image_pair();


	/**
	 * @brief Beregner sammenlikningsgrunnlaget for mapping av regioner i bildeparet.
	 *
	 * @param weight Peker til en vektor som inneholder vektingen av den lille 
	 *				 egenskapsmengden (areal, massesenter og omlsuttende rektangel).
	 *
	 * @return Peker til vektor som inneholder sammenlikningsgrunnlaget for
	 *		   mapping av regioner i bildeparet.
	 */
	int * calculateComparisonMatrix(float *weight);


	/**
	 * @brief Beregner beste korrespondans/mapping mellom regioner i bildeparet.
	 *
	 * @param comparison Peker til vektor som inneholder sammenlikningsgrunnlaget
	 *					 for mapping av regioner i bildeparet.
	 * @param sort Peker til sorteringsobjektet sort.
	 */
	void calculateBestCorrespondenceVector(int *comparison, Sort *sort);


	/**
	 * @brief Henter den beste korrespondans/mapping mellom regioner i bildeparet.
	 *
	 * @return Peker til vektor som inneholder den beste mappingen/korrespondansen
	 *		   mellom regioner i bildeparet.
	 */
	int * getBestCorrespondenceVector();


	/**
	 * @brief Henter det bildet av bildeparet som har minst antall regioner.
	 *
	 * @return Peker til det bildet av bildeparet som har minst antall regioner.
	 */
	Image * getSmallImage();


	/**
	 * @brief Henter det bildet av bildeparet som har strst antall regioner.
	 *
	 * @return Peker til det bildet av bildeparet som har strst antall regioner.
	 */
	Image * getBigImage();


	/**
	 * @brief Henter ut informasjon om bildeparet er homogent eller ikke.
	 *
	 * @return True dersom bildeparet bestr av to godkjente bilder, false ellers. 
	 */
	bool getHomogenious();

};
