
//Refereanser til eksternt grensesnitt/kode som anvendes i denne filen
#include "Region.h"


/**
 * @brief Konstruktor
 *
 * @param _pixels Dynamisk vektor som holder pikslene i denne regionen.
 * @param _border Dynamisk vektor som holder de piksler som er kontur i denne regionen.
 * @param _r Variabel som holder regionens gjennomsnittlige rde verdi.
 * @param _g Variabel som holder regionens gjennomsnittlige grnnne verdi.
 * @param _b Variabel som holder regionens gjennomsnittlige bl verdi.
 */
Region::Region( vector<Pixel> &_pixels, vector<Pixel> &_border, int _r, int _g, int _b)
{
	pixels = _pixels;
	border = _border;
	calculated = false;
	r = _r / pixels.size();
	g = _g / pixels.size();
	b = _b / pixels.size();
}



/**
 * @brief Destruktor
 */
Region::~Region()
{
	//Tom destruktor
}



/**
 * @brief Beregner hvilken piksel som tilsvarer regionens massesenterpunkt.
 */
void Region::calculateCentroid()
{
	unsigned int x=0, y=0;
	for (unsigned int i=0;i<pixels.size();i++)
	{
		x += pixels.at(i).x;
		y += pixels.at(i).y;
	}
	x /= pixels.size();
	y /= pixels.size();
	Pixel p(x,y,0,0,0);
	centroid = p;
}



/**
 * @brief Beregner regionens ekstremalverdier/pikselposisjoner i nord, st, sr og vest.
 */
void Region::calculateBoundingBox()
{
	int unsigned n=pixels.size(), e=0, s=0, w=pixels.size();
	for (unsigned int i=0;i<pixels.size();i++)
	{
		if ( pixels.at(i).y < n ) { n = pixels.at(i).y; }
		if ( pixels.at(i).x > e ) { e = pixels.at(i).x; }
		if ( pixels.at(i).y > s ) { s = pixels.at(i).y; }
		if ( pixels.at(i).x < w ) { w = pixels.at(i).x; }
	}
	boundingBox[0] = n;
	boundingBox[1] = e;
	boundingBox[2] = s;
	boundingBox[3] = w;
}



/**
 * @brief Setter hvilket vindu denne regionen hrer til i bildet.
 *
 * @param window Vindu som betegner aktuell posisjon i bildet.
 */
void Region::setWindow(int _window)
{
	window = _window;
}



/**
 * @brief Henter regionens massesenterpunkt.
 *
 * @return Piksel som tilsvarer regionens massesenterpunkt.
 */
Pixel Region::getCentroid()
{
	return centroid;
}



/**
 * @brief Henter regionens areal.
 *
 * @return Antall piskler i denne regionen.
 */
int Region::getArea()
{
	return pixels.size();
}



/**
 * @brief Henter regionens ekstremalverdier.
 *
 * @return Peker til vektor som holder piksler tilsvarende regionens ytterpunkter i nord, st, sr og vest.
 */
int * Region::getBoundingBox()
{
	return boundingBox;
}



/**
 * @brief Henter det vindu som denne regionen hrer til.
 *
 * @return Vinduet denne regionen tilhrer.
 */
int Region::getWindow()
{
	return window;
}



/**
 * @brief Henter pikslene i denne regionen.
 * 
 * @return En dynamisk vektor som holder pikslene i denne regionen.
 */
vector<Pixel> Region::getPixels()
{
	return pixels;
}



/**
 * @brief Henter de piskler som ligger p konturen av denne regionen.
 * 
 * @return En dynamisk vektor som holder de piksler som ligger p denne regionens kontur.
 */
vector<Pixel> Region::getBorder()
{
	return border;
}



/**
 * @brief Merker av at egenskapskandidatberegninger for denne regionen er utfrt.
 */
void Region::setCalculated()
{
	calculated = true;
}



/**
 * @brief Henter ut informasjon om egenskapskandidater er beregner for denne regionen.
 *
 * @return True derson egenskapskandidater er beregnet for denne regionen, false ellers.
 */
bool Region::getCalculated()
{
	return calculated;
}



/**
 * @brief Setter dimensjon til vektor som holder beregnede verdier for egenskapskandidater for denne regionen. 
 *
 * @param numberOfFeatureCandidates Antall egenskapskandidater.
 */
void Region::setFeatureVectorRowSize(int numberOfFeatureCandidates)
{
	featureVector = new float*[numberOfFeatureCandidates];
}



/**
 * @brief Registrerer verdier for beregnede egenskapskandidaters.
 *
 * @param row Angir aktuell egenskapskandidat (rad i featureVector).
 * @param cols Angir dimensjonen til aktuell egenskapskandidat (vektordimensjon i featureVector[rad]).
 * @param values Peker til beregnede verdier for aktuell egenskapskandidat for denne regionen.
 */
void Region::setFeatureVectorValue(int row, int cols, float *values)
{
	featureVector[row] = new float[cols];
	for (int i=0;i<cols;i++)
	{
		featureVector[row][i] = values[i];
	}
}



/**
 * @brief Henter en komponent av egenskapskandidatens beregnede verdi av denne regionen.
 *
 * @param row Angir aktuell egenskapskandidat.
 * @param col Angir aktuell komponent av egenskapskandidatens dimensjon.
 *
 * @return Flytverdi tilsvarende en komponent av egenskapskandidatens beregnede verdi av denne regionen.
 */
float Region::getFeatureValue(int row, int col)
{
	return featureVector[row][col];
}



/**
 * @brief Henter regionens gjennomsnittlige rde verdi.
 *
 * @return Regionens gjennomsnittlige rde verdi.
 */
int Region::getR()
{
	return r;
}



/**
 * @brief Henter regionens gjennomsnittlige grnne verdi.
 *
 * @return Regionens gjennomsnittlige grnne verdi.
 */
int Region::getG()
{
	return g;
}



/**
 * @brief Henter regionens gjennomsnittlige bl verdi.
 *
 * @return Regionens gjennomsnittlige bl verdi.
 */
int Region::getB()
{
	return b;
}