#pragma once

//Refereanser til eksternt grensesnitt/kode som anvendes i denne filen
#include <vector>
#include "Pixel.h"

//Denne kilden bruker kode definert i navnerommet std
using namespace std;


/**
 * @class Region
 * @brief Klasse som representerer regionobjekter i et bilde.
 *
 * @version 1.0
 * @author Per Gunnar Brdsen
 */
class Region
{

	///Dynamisk vektor som holder pikslene i denne regionen.
	vector<Pixel> pixels;

	///Dynamisk vektor som holder de piksler som er kontur i denne regionen.
	vector<Pixel> border;

	///Piksel som tilsvarer massesenter i denne regionen.
	Pixel centroid;

	///Vektor som holder pikselposisjoner til denne regionenes ekstremalverdier i nord, st, sr og vest.
	int boundingBox [4];

	///Variabel som holder hvilket vindu denne regionen tilsvarer i bildet.
	int window;

	///Variabel som forteller om det er beregnet egenskapskandidater for regionen eller ikke.
	bool calculated;
	
	///Vektor (v) av vektorer (w), der v har en rad for hver egenskapskandidat, og w bestr av 
	///denne regionens egenskapskandidatberegninger for egenskapskandidaten.
	float **featureVector;

	///Variabel som holder regionens gjennomsnittlige rde verdi.
	int r;
	
	///Variabel som holder regionens gjennomsnittlige grnnne verdi.
	int g;
	
	///Variabel som holder regionens gjennomsnittlige bl verdi.
	int b;

public:

	/**
	 * @brief Konstruktor
	 *
	 * @param _pixels Dynamisk vektor som holder pikslene i denne regionen.
	 * @param _border Dynamisk vektor som holder de piksler som er kontur i denne regionen.
	 * @param _r Variabel som holder regionens gjennomsnittlige rde verdi.
	 * @param _g Variabel som holder regionens gjennomsnittlige grnnne verdi.
	 * @param _b Variabel som holder regionens gjennomsnittlige bl verdi.
	 */
	Region(vector<Pixel> &_pixels, vector<Pixel> &_border, int _r, int _g, int _b);


	/**
	 * @brief Destruktor
	 */
	~Region();


	/**
	 * @brief Beregner hvilken piksel som tilsvarer regionens massesenterpunkt.
	 */
	void calculateCentroid();


	/**
	 * @brief Beregner regionens ekstremalverdier/pikselposisjoner i nord, st, sr og vest.
	 */
	void calculateBoundingBox();


	/**
	 * @brief Setter hvilket vindu denne regionen hrer til i bildet.
	 *
	 * @param window Vindu som betegner aktuell posisjon i bildet.
	 */
	void setWindow(int window);


	/**
	 * @brief Henter regionens massesenterpunkt.
	 *
	 * @return Piksel som tilsvarer regionens massesenterpunkt.
	 */
	Pixel getCentroid();


	/**
	 * @brief Henter regionens areal.
	 *
	 * @return Antall piskler i denne regionen.
	 */
	int getArea();



	/**
	 * @brief Henter regionens ekstremalverdier.
	 *
	 * @return Peker til vektor som holder piksler tilsvarende regionens ytterpunkter i nord, st, sr og vest.
	 */
	int * getBoundingBox();



	/**
	 * @brief Henter det vindu som denne regionen hrer til.
	 *
	 * @return Vinduet denne regionen tilhrer.
	 */
	int getWindow();



	/**
	 * @brief Henter pikslene i denne regionen.
	 * 
	 * @return En dynamisk vektor som holder pikslene i denne regionen.
	 */
	vector<Pixel> getPixels();


	/**
	 * @brief Henter de piskler som ligger p konturen av denne regionen.
	 * 
	 * @return En dynamisk vektor som holder de piksler som ligger p denne regionens kontur.
	 */
	vector<Pixel> getBorder();



	/**
	 * @brief Merker av at egenskapskandidatberegninger for denne regionen er utfrt.
	 */
	void setCalculated();



	/**
	 * @brief Henter ut informasjon om egenskapskandidater er beregner for denne regionen.
	 *
	 * @return True derson egenskapskandidater er beregnet for denne regionen, false ellers.
	 */
	bool getCalculated();


	/**
	 * @brief Setter dimensjon til vektor som holder beregnede verdier for egenskapskandidater for denne regionen. 
	 *
	 * @param numberOfFeatureCandidates Antall egenskapskandidater.
	 */
	void setFeatureVectorRowSize(int numberOfFeatureCandidates);


	/**
	 * @brief Registrerer verdier for beregnede egenskapskandidater.
	 *
	 * @param row Angir aktuell egenskapskandidat (rad i featureVector).
	 * @param cols Angir dimensjonen til aktuell egenskapskandidat (vektordimensjon i featureVector[rad]).
	 * @param values Peker til beregnede verdier for aktuell egenskapskandidat for denne regionen.
	 */
	void setFeatureVectorValue(int row, int cols, float *values);



	/**
	 * @brief Henter en komponent av egenskapskandidatens beregnede verdi av denne regionen.
	 *
	 * @param row Angir aktuell egenskapskandidat.
	 * @param col Angir aktuell komponent av egenskapskandidatens dimensjon.
	 *
	 * @return Flytverdi tilsvarende en komponent av egenskapskandidatens beregnede verdi av denne regionen.
	 */
	float getFeatureValue(int row, int col);



	/**
	 * @brief Henter regionens gjennomsnittlige rde verdi.
	 *
	 * @return Regionens gjennomsnittlige rde verdi.
	 */
	int getR();


	/**
	 * @brief Henter regionens gjennomsnittlige grnne verdi.
	 *
	 * @return Regionens gjennomsnittlige grnne verdi.
	 */
	int getG();


	/**
	 * @brief Henter regionens gjennomsnittlige bl verdi.
	 *
	 * @return Regionens gjennomsnittlige bl verdi.
	 */
	int getB();

};